#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



from netq_lib.orm.redisdb.models import Ntp


def process_ntpq(output, **kwargs):
    ''' Parse ntpq -p command output
         - Function extracts active server information
         - If activer server is found it returns ntp_sync = 'yes'
    '''
    current_server = '-'
    stratum = 16
    ntp = Ntp()
    ntp_sync = 'no'

    if output:
        for line in output.splitlines():
            columns = line.split()
            if len(columns) < 3:
                continue
            if '*' in columns[0]:
                current_server = columns[0][1:]
                if len(columns) >= 8:
                    stratum = columns[-8]
                    ntp_sync = 'yes'
                else:
                    stratum = 16

    ntp.current_server = current_server
    ntp.stratum = stratum
    ntp.ntp_sync = ntp_sync
    ntp.ntp_app = 'ntpq'
    ntp.save()

    return ntp_sync

def process_timedatectl(output, **kwargs):
    ''' Parse timedatectl status command output
         - checks for 'Network time on: yes' or 'NTP synchronized: yes'
         - checks for 'systemd-timesyncd.service active: yes' (ub1804)
         - If found, returns ntp_sync = 'yes'
         - verified on ub1604 w/ systemd v229
         - verified on ub1804 w/ systemd v237
    '''
    ntp = Ntp()
    ntp_sync = 'no'

    if output:
        for line in output.splitlines():
            columns = line.split(':')
            # ubuntu and debian datetimectl format differences
            if 'Network time on' in columns[0] or \
                    'NTP synchronized' in columns[0] or \
                    'systemd-timesyncd.service active' in columns[0]:
                ntp_sync = columns[1].lstrip()
                break

    # current server and stratum not known
    # when using timedatectl / systemd-timesyncd
    ntp.current_server = '-'
    ntp.stratum = 16
    ntp.ntp_sync = ntp_sync
    ntp.ntp_app = 'timedatectl'
    ntp.save()

    return ntp_sync

def process_chronyc(output, **kwargs):
    ''' Parse chronyc sources command output
         - Function extracts active server information
         - If active server is found it returns ntp_sync = 'yes'
         - Eg.
# chronyc  sources
210 Number of sources = 4
MS Name/IP address         Stratum Poll Reach LastRx Last sample
===============================================================================
^+ up2.com                       2  10   377   733   -815us[ -815us] +/-   95ms
^+ ntp1.csl.tjhsst.edu           3   8   377   228   +239us[ +239us] +/-   89ms
^? excalibur.prolixium.com       0  10     0     -     +0ns[   +0ns] +/-    0ns
^* resolver1.skyfiberintern>     2  10   377   939   +277us[ +404us] +/-   44ms

    '''
    current_server = '-'
    stratum = 16
    ntp = Ntp()
    ntp_sync = 'no'

    if output:
        for line in output.splitlines():
            columns = line.split()
            if len(columns) < 3:
                continue
            if '*' in columns[0]:
                current_server = columns[1]
                if len(columns) >= 2:
                    stratum = columns[2]
                    ntp_sync = 'yes'
                else:
                    stratum = 16

    ntp.current_server = current_server
    ntp.stratum = stratum
    ntp.ntp_sync = ntp_sync
    ntp.ntp_app = 'chronyc'
    ntp.save()

    return ntp_sync
