#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import yaml
import ujson as json
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Roce, RoceConfig
from netq_lib.common.utils import save_fullstate, subprocess_wrapper, check_cl_version, get_cl_version, is_active_systemd_svc
from netq_lib.common.enums import Platform, ServiceStatus, DefaultPaths
from netq_agent.parsers.common.render import BaseRender
from netq_lib.common import configuration


class RoceRender(BaseRender):
    def __init__(self):
        super(RoceRender, self).__init__()
        self.port_id_name_mapping = dict()
        self.last_roce_config_modify_time = None
        self.switchd_last_reset_time = None
        self.is_roce_new_ver = self.is_roce_new_version()

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')
        if not configuration.is_cl4():
            return
        if Platform.MELLANOX not in platform:
            return

        if os.path.exists(DefaultPaths.ROCE_DUMMY_DATA_PATH):
            try:
                with open(DefaultPaths.ROCE_DUMMY_DATA_PATH, 'r') as f:
                    test_roce_data = yaml.safe_load(f)

                    roce_intf_counters = test_roce_data.get('roce_counters', [])
                    for intf_counter in roce_intf_counters:
                        roce_obj = Roce()
                        roce_obj_attr = dir(roce_obj)
                        for ifname, intf_counters in list(intf_counter.items()):
                            setattr(roce_obj, 'ifname', ifname)
                            cell_unit_size = intf_counters.get('cell_unit_size', 96)
                            for key, value in list(intf_counters.items()):
                                if key in roce_obj_attr:
                                    if 'usage_cells' in key:
                                        setattr(roce_obj, key, value * cell_unit_size)
                                    else:
                                        setattr(roce_obj, key, value)
                        roce_obj.save()

            except Exception as e:
                pass
            return

        if not self.is_roce_enabled():
            self.delete_all_roce_counters()
            return

        pool_size_mapping = self.roce_pool_size_json() if self.is_roce_new_ver else self.roce_pool_size()

        try:
            cmd_json_output = json.loads(cmd_output)
        except ValueError:
            self.delete_all_roce_counters()
            return ServiceStatus.FAIL

        fullstate_arr = []
        if 'exception' in cmd_json_output:
            return

        for per_intf_dict in list(cmd_json_output):
            roce_obj = Roce()
            roce_obj_attr = dir(roce_obj)
            cell_unit_size = per_intf_dict.get('cell_unit_size', 96)
            for key, value in list(per_intf_dict.items()):
                if value is None:
                    continue
                if key in roce_obj_attr:
                    if 'usage_cells' in key:
                        setattr(roce_obj, key, value * cell_unit_size)
                    else:
                        setattr(roce_obj, key, value)
                if key == 'port_id':
                    ifname = self._get_ifname_from_port_id(value)
                    setattr(roce_obj, 'ifname', ifname)

            for key, value in list(pool_size_mapping.items()):
                if key in roce_obj_attr:
                    setattr(roce_obj, key, value)

            fullstate_arr.append(roce_obj)
            if not fullstate:
                roce_obj.save()

        if fullstate:
            save_fullstate(fullstate_arr, Roce().__class__.__name__)

    def is_roce_enabled(self):
        roce_enabled = False
        if os.path.exists(DefaultPaths.ROCE_STATUS_FILE_PATH):
            try:
                with open(DefaultPaths.ROCE_STATUS_FILE_PATH, 'r') as fd:
                    for line in fd.readlines():
                        if 'traffic.roce_mode' in line:
                            mode = int(line.strip().split('=')[1].strip())
                            if mode != 0:
                                roce_enabled = True
                            break
            except Exception as e:
                pass
        return roce_enabled


    def is_roce_new_version(self):
        try:
            version = get_cl_version()
            if version.startswith('4.4') or version.startswith('5'):
                return True
            else:
                return False
        except Exception:
            return False


    def is_roce_config_changed(self):
        switchd_reset = False
        try:
            cmd_out = subprocess_wrapper('/bin/systemctl show -p StateChangeTimestamp switchd'.split())
            state_change_time = cmd_out.splitlines()[0].split('=')[1]
            if self.switchd_last_reset_time is None or state_change_time != self.switchd_last_reset_time:
                self.switchd_last_reset_time = state_change_time
                switchd_reset = True
        except ValueError:
            pass

        return self.is_config_changed(DefaultPaths.TRAFFIC_CONF_FILE_PATH) or \
               self.is_config_changed(DefaultPaths.QOS_CONF_FILE_PATH) or switchd_reset

    def is_config_changed(self, roce_config_file_path):
        roce_config_changed = False
        if os.path.exists(roce_config_file_path):
            try:
                current_modify_time = int(os.path.getmtime(roce_config_file_path))
                if self.last_roce_config_modify_time is None:
                    roce_config_changed = True
                elif current_modify_time > self.last_roce_config_modify_time:
                    roce_config_changed = True

                if roce_config_changed:
                    self.last_roce_config_modify_time = current_modify_time
            except Exception as e:
                pass
        return roce_config_changed

    def parse_roce_config(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')
        cmd_json_output = []

        if os.path.exists(DefaultPaths.ROCECONFIG_DUMMY_DATA_PATH):
            try:
                with open(DefaultPaths.ROCECONFIG_DUMMY_DATA_PATH, 'r') as f:
                    cmd_json_output = yaml.safe_load(f)
            except Exception as e:
                pass

        else:
            if (Platform.MELLANOX not in platform) or (not configuration.is_cl4()):
                return

            if not self.is_roce_enabled():
                self.delete_all_roceconfig()
                return

            if not self.is_roce_config_changed() and not fullstate:
                return

            if os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    self.delete_all_roceconfig()
                    return ServiceStatus.FAIL

            try:
                roce_status_output = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json roce status'.split())
                cmd_json_output = json.loads(roce_status_output)
            except ValueError:
                self.delete_all_roceconfig()
                return ServiceStatus.FAIL

        congestion_control_from_qos = None
        ets_priority_mode = None

        if 'exception' in cmd_json_output:
            self.delete_all_roceconfig()
            return

        m_roceconfig = RoceConfig.memmodel()
        old_config_objs = []
        new_config_objs = []
        if m_roceconfig:
            old_config_objs = (
                list(m_roceconfig.query.filter(hostname=HOSTNAME))
            )

        for roce_status in list(cmd_json_output):
            roce_config_obj = RoceConfig()

            if 'congestion-control' in roce_status:
                congestion_control = roce_status.get('congestion-control')
                roce_config_obj.ecn_min_th = congestion_control.get('min_threshold_bytes', 0)
                roce_config_obj.ecn_max_th = congestion_control.get('max_threshold_bytes', 0)
                ecn_tcs = congestion_control.get('ecn_tcs')
                roce_config_obj.enabled_tc = ','.join(str(x) for x in ecn_tcs)
                ecn_enabled = congestion_control.get('ecn_enabled', 0) == 1
                red_enabled = congestion_control.get('red_enabled', 0) == 1

                mode = 'None'
                if ecn_enabled and not red_enabled:
                    mode = 'ECN'
                elif not ecn_enabled and red_enabled:
                    mode = 'RED'
                elif ecn_enabled and red_enabled:
                    mode = 'ECN+RED'
                roce_config_obj.mode = mode

            else:
                if congestion_control_from_qos is None:
                    congestion_control_from_qos = self.get_congestion_control()

                roce_config_obj.enabled_tc = congestion_control_from_qos.get('enabled_tc', 'None')
                roce_config_obj.mode = congestion_control_from_qos.get('mode', 'None')
                roce_config_obj.ecn_min_th  = congestion_control_from_qos.get('min_th', 0)
                roce_config_obj.ecn_max_th  = congestion_control_from_qos.get('max_th', 0)

            if 'roce_dscp' in roce_status:
                roce_config_obj.dscp_sp_map = roce_status.get('roce_dscp',{})
            else:
                roce_config_obj.dscp_sp_map = roce_status.get('dscp',{})

            if 'roce_pg_map' in roce_status:
                roce_config_obj.sp_pg_map = roce_status.get('roce_pg_map',{})
            else:
                roce_config_obj.sp_pg_map = roce_status.get('pg_map',{})

            if 'roce_tc_map' in roce_status:
                roce_config_obj.sp_tc_map = roce_status.get('roce_tc_map',{})
            else:
                roce_config_obj.sp_tc_map = roce_status.get('tc_map',{})

            roce_config_obj.cnp_dscp = roce_status.get('cnp_dscp',{})
            roce_config_obj.cnp_pg_map = roce_status.get('cnp_pg_map',{})
            roce_config_obj.cnp_tc_map = roce_status.get('cnp_tc_map',{})
            roce_config_obj.roce_mode = roce_status.get('roce_mode')

            if 'port' in roce_status:
                roce_config_obj.ifname = roce_status.get('port')
            else:
                port_id = roce_status.get('port_id')
                roce_config_obj.ifname = self._get_ifname_from_port_id(port_id)

            rx_enabled = roce_status.get('pfc_enabled', {}).get('rx_enabled', 'no') == 'yes'
            tx_enabled = roce_status.get('pfc_enabled', {}).get('tx_enabled', 'no') == 'yes'
            roce_config_obj.pfc_rx_enabled = rx_enabled
            roce_config_obj.pfc_tx_enabled = tx_enabled
            roce_config_obj.pfc_enabled_sp = roce_status.get('pfc_priority') if rx_enabled and tx_enabled else '-'

            if 'cnp_sched_map' in roce_status:
                roce_config_obj.cnp_ets_mode =  roce_status.get('cnp_sched_map', {}).get('dwrr_mode')

            if 'roce_sched_map' in roce_status:
                roce_config_obj.ets_mode =  roce_status.get('roce_sched_map', {}).get('dwrr_mode')
            else:
                if ets_priority_mode is None:
                    ets_priority_mode = self.get_ets_priority_mode()
                roce_config_obj.ets_mode = ets_priority_mode

            new_config_objs.append(roce_config_obj)

        # remove current elements from old list
        for new_ele in new_config_objs:
            for old_ele in old_config_objs:
                if old_ele.ifname == new_ele.ifname:
                    old_config_objs.remove(old_ele)

        # delete remaining stale entries
        for old_ele in old_config_objs:
            old_config_ent = RoceConfig()
            old_config_ent.copy(old_ele)
            old_config_ent.delete()

        for roce_config_obj in new_config_objs:
            roce_config_obj.save()

        if fullstate:
            save_fullstate(new_config_objs, RoceConfig().__class__.__name__)

    def delete_all_roceconfig(self):
        m_roceconfig = RoceConfig.memmodel()
        if m_roceconfig:
            for entry in m_roceconfig.query.filter(hostname=HOSTNAME):
                old_roce_config = RoceConfig()
                old_roce_config.copy(entry)
                old_roce_config.delete()

    def delete_all_roce_counters(self):
        m_roce = Roce.memmodel()
        if m_roce:
            for entry in m_roce.query.filter(hostname=HOSTNAME):
                old_roce = Roce()
                old_roce.copy(entry)
                old_roce.delete()

    def _get_ifname_from_port_id(self, port_id):
        if str(port_id) not in self.port_id_name_mapping:
            self.port_id_name_mapping = self.roce_port_info()
            print("Refreshed Port Id to Interface name mapping : {}".format(self.port_id_name_mapping))
        return self.port_id_name_mapping.get(str(port_id))
